#include <GL/gl.h>
#include <libguile.h>
#include <guile-gtk.h>
#include "config.h"

/*
 *
 *	functions too awkward to be in GL*.defs
 */

static SCM sgl_local_protector = SCM_BOOL_F;
extern sgtk_enum_info sgtk_gl_feedback_type_info;

static void 
sgl_init_protector (SCM *s_prot)
{
  *s_prot = scm_cons (SCM_EOL, SCM_EOL);
#if defined (HAVE_SCM_GC_PROTECT_OBJECT)
  scm_gc_protect_object (*s_prot);
#else
  scm_protect_object (*s_prot);
#endif
}

static void
sgl_protect (SCM s_protector, SCM object)
{
  SCM    newc;

  newc = scm_cons (object, SCM_CDR (s_protector));
  SCM_SETCDR (s_protector, newc);
}

static void
sgl_unprotect (SCM s_protector, SCM object)
{
  SCM		list, previous;

  previous = s_protector;
  list = SCM_CDR (previous);

  while (list != SCM_EOL)
    {
      if (SCM_CAR (list) == object)
	break;

      previous = list;
      list = SCM_CDR (list);
    }

  if (list != SCM_EOL)
    SCM_SETCDR (previous, SCM_CDR (list));
}

SCM_PROC (s_gl_select_buffer, "gl-select-buffer", 1, 0, 0, sgl_select_buffer);

static SCM
sgl_select_buffer (SCM s_buffer)
{
  static SCM	sel_buffer = SCM_BOOL_F;

  SCM_ASSERT (SCM_NIMP (s_buffer) && (SCM_TYP7 (s_buffer) == scm_tc7_uvect),
	      s_buffer, SCM_ARG1, s_gl_select_buffer);

  sgl_unprotect (sgl_local_protector, sel_buffer);
  sgl_protect (sgl_local_protector, sel_buffer = s_buffer);

  glSelectBuffer (SCM_LENGTH (s_buffer), (unsigned int *) SCM_VELTS (s_buffer));

  return SCM_UNSPECIFIED;  
}

SCM_PROC (s_gl_feedback_buffer, "gl-feedback-buffer", 2, 0, 0, sgl_feedback_buffer);

static SCM
sgl_feedback_buffer (SCM s_type, SCM s_buffer)
{
  static SCM	feedback_buffer = SCM_BOOL_F;
  int		c_type;

  SCM_ASSERT (SCM_NIMP (s_buffer) && (SCM_TYP7 (s_buffer) == scm_tc7_fvect),
	      s_buffer, SCM_ARG2, s_gl_feedback_buffer);

  c_type = sgtk_scm2enum (s_type, &sgtk_gl_feedback_type_info, SCM_ARG1, (char *) sgl_feedback_buffer);

  sgl_unprotect (sgl_local_protector, feedback_buffer);
  sgl_protect (sgl_local_protector, feedback_buffer = s_buffer);

  glFeedbackBuffer (SCM_LENGTH (s_buffer), c_type,  (float *) (SCM_VELTS (s_buffer)));

  return SCM_UNSPECIFIED;  
}

void sgl_init_GL_support_glue (void)
{
  sgl_init_protector (& sgl_local_protector);

#if !(defined(SCM_MAGIC_SNARFER) || defined(SCM_MAGIC_SNARF_INITS))
#include "GL-support.x"
#endif
}

/*
 *     ext_arrays are small read only arrays of floats or symbols (enums)
 *     Usually they do not exceed length of four floats (so we expect
 *     that they are allocated on stack * and we do not bother to
 *     allocate them using malloc 
*/

#define MMIN(A,B)		(((A) < (B)) ? (A) : (B))

static int
sgl_fext_member (SCM obj, GLfloat *result, sgtk_enum_info *info)
{
  if (SCM_INUMP (obj))
    { *result = (GLfloat) SCM_INUM (obj); return TRUE; }

  if (SCM_SYMBOLP (obj))
    { 
      int tmp;

      if (sgtk_enum_flags_bin_search (obj, info, &tmp) == FALSE)
	return FALSE;

      *result = (GLfloat) tmp;
      return TRUE;
    }

  if (scm_number_p (obj) == SCM_BOOL_T)
    {
      *result = (GLfloat) scm_num2dbl (obj, "");
      return TRUE;
    }

  return FALSE;
}

GLfloat *
sgl_scm2fext_array (SCM obj, GLfloat *array, int maxsize, sgtk_enum_info *info,
		int pos, char *sname)
{
  if (SCM_IMP (obj))
    {
      if (sgl_fext_member (obj, array, info))
	return array;

      goto error;
    }

  if (SCM_TYP7 (obj) == scm_tc7_fvect)
    return (GLfloat *) SCM_VELTS(obj);
  else
    if (SCM_CONSP (obj))
      {			/* list */
	SCM	data;
	GLfloat *ptr;

	ptr = array;

	while (obj != SCM_EOL)
	  {
	    if ((maxsize--) == 0)
	      goto error;

	    if (! SCM_CONSP (obj))
	      goto error;
	    
	    data = SCM_CAR (obj);

	    if (! sgl_fext_member (data, ptr, info))
	      goto error;

	    ptr++;
	    obj = SCM_CDR (obj);
	  }

	return array;
      }
    else
      if (SCM_VECTORP (obj))
	{
	  GLfloat *ptr;
	  SCM	   *scmptr;
	  int	   i, j;

	  j = MMIN ((int) SCM_LENGTH (obj), maxsize);
	  scmptr = (SCM *) SCM_VELTS (obj);
	  ptr = array;

	  for (i = 0 ; i < j; i++, ptr++, scmptr++)
	    if (! sgl_fext_member (*scmptr, ptr, info))
	      goto error;

	  return array;
	}
 error:
  SCM_ASSERT (0, obj, pos, sname);
  return NULL;
}

/*
 *     x-arrays are small read only arrays of x (x is float, double...)
 *     Usually they do not exceed length of four floats (so we expect
 *     that they are allocated on stack * and we do not bother to
 *     allocate them using malloc 
 *     (largest size of f-array is 16 * sizeof(float) - in glMultMatrixf
 */

#define SGL_SCM2X_ARRAY(NAME, TYPE, TC7)							\
												\
TYPE *												\
sgl_scm2 ## NAME ## _array (SCM obj, TYPE *array, int maxsize, int exact_size,			\
			    int pos, const char *sname)						\
{												\
  if (SCM_IMP (obj))										\
    goto error;											\
												\
  if (SCM_TYP7 (obj) == TC7)									\
    {												\
      if (exact_size && (((int) SCM_LENGTH (obj)) != maxsize))					\
	goto error;										\
      return (TYPE *) SCM_VELTS(obj);								\
    }												\
  else												\
    if (SCM_CONSP (obj))									\
      {			/* list */								\
	SCM	data;										\
	TYPE    *ptr;										\
												\
	ptr = array;										\
												\
	while (obj != SCM_EOL)									\
	  {											\
	    if ((maxsize--) == 0)								\
	      goto error;									\
												\
	    if (! SCM_CONSP (obj))								\
	      goto error;									\
	    											\
	    data = SCM_CAR (obj);								\
												\
	    if (! IS_A (data))									\
	      goto error;									\
												\
	    *ptr++ = SCM2C (data);								\
												\
	    obj = SCM_CDR (obj);								\
	  }											\
												\
	if (exact_size && (maxsize != 0))							\
	  goto error;										\
												\
	return array;										\
      }												\
    else											\
      if (SCM_VECTORP (obj))									\
	{											\
	  TYPE	   *ptr;									\
	  SCM	   *scmptr;									\
	  int	   i, j;									\
												\
	  if (exact_size && (((int)SCM_LENGTH (obj)) != maxsize))				\
	    goto error;										\
												\
	  j = MMIN ((int) SCM_LENGTH (obj), maxsize);						\
	  scmptr = (SCM *) SCM_VELTS (obj);							\
	  ptr = array;										\
												\
	  for (i = 0 ; i < j; i++, ptr++, scmptr++)						\
	    {											\
	      if (! IS_A (*scmptr))								\
		goto error;									\
												\
	      *ptr = SCM2C (*scmptr);								\
	    }											\
												\
	  return array;										\
	}											\
 error:												\
  SCM_ASSERT (0, obj, pos, sname);								\
  return NULL;											\
}

/*
 *
 *	GLfloat - array
 */

#define IS_A(X)		(scm_number_p (X) == SCM_BOOL_T)
#define SCM2C(X)	((GLfloat) (scm_num2dbl (X, "")))

SGL_SCM2X_ARRAY (f, GLfloat, scm_tc7_fvect)

#undef SCM2C
#undef IS_A

/*
 *
 *	GLdouble - array
 */

#define IS_A(X)		(scm_number_p (X) == SCM_BOOL_T)
#define SCM2C(X)	((GLdouble) (scm_num2dbl (X, "")))

SGL_SCM2X_ARRAY (d, GLdouble, scm_tc7_dvect)

#undef SCM2C
#undef IS_A

/*
 *
 *	GLint - array
 */

#define IS_A(X)		(scm_integer_p (X) == SCM_BOOL_T)
#define SCM2C(X)	((GLint) (scm_num2long (X, 0, "")))

SGL_SCM2X_ARRAY (i, GLint, scm_tc7_ivect)

#undef SCM2C
#undef IS_A

/*
 *
 *	GLuint - array
 */

#define IS_A(X)		(scm_integer_p (X) == SCM_BOOL_T)
#define SCM2C(X)	((GLuint) (scm_num2long (X, 0, "")))

SGL_SCM2X_ARRAY (u, GLuint, scm_tc7_uvect)

#undef SCM2C
#undef IS_A

/*
 *
 *	GLshort - array
 */

#define IS_A(X)		(scm_integer_p (X) == SCM_BOOL_T)
#define SCM2C(X)	((GLint) (scm_num2long (X, 0, "")))

SGL_SCM2X_ARRAY (s, GLshort, scm_tc7_svect)

#undef SCM2C
#undef IS_A


/*
 *
 *	GLbyte - array
 */

#define IS_A(X)		(scm_integer_p (X) == SCM_BOOL_T)
#define SCM2C(X)	((GLbyte) (scm_num2long (X, 0, "")))

SGL_SCM2X_ARRAY (b, GLubyte, scm_tc7_byvect)

#undef SCM2C
#undef IS_A

/*
 *
 *	GLubyte - array
 */

#define IS_A(X)		(scm_integer_p (X) == SCM_BOOL_T)
#define SCM2C(X)	((GLubyte) (scm_num2long (X, 0, "")))

SGL_SCM2X_ARRAY (ub, GLbyte, scm_tc7_string)

#undef SCM2C
#undef IS_A
